#include <lektor/common.h>
#include <lektor/json.h>

#define __JSON_SPACE " \n\t:=,"
#define __JSON_BEGIN "{["
#define __JSON_END   "}]"
#define __JSON_SEP   '"'

#define __SKIP_JSON(str) (str = &str[strspn(str, __JSON_SPACE)])

#define __NEXT_JSON(str, len, dest, tmp, tmp_len)                               \
    {                                                                           \
        int is_paren = 0;                                                       \
        if (str[0] == __JSON_SEP) {                                             \
            ++str;                                                              \
            const char *begin = str, *end;                                      \
            for (;;) {                                                          \
                end = strchr(begin, __JSON_SEP);                                \
                if (*(end - 1) == '\\') {                                       \
                    begin = end + 1;                                            \
                    continue;                                                   \
                }                                                               \
                break;                                                          \
            }                                                                   \
            len      = (size_t)(end - str);                                     \
            is_paren = 1;                                                       \
        } else                                                                  \
            len = strcspn(str, __JSON_SPACE __JSON_END);                        \
        if (level == asked_level) {                                             \
            size_t local_len = LKT_LINE_MAX - 1 < len ? LKT_LINE_MAX - 1 : len; \
            strncpy(dest, str, local_len);                                      \
            __decode_json(dest, local_len);                                     \
        }                                                                       \
        str += len;                                                             \
        if (is_paren) {                                                         \
            ++str;                                                              \
        }                                                                       \
    }

#define __SKIP_NEXT_JSON(str, len)                                    \
    {                                                                 \
        int is_paren = 0;                                             \
        if (str[0] == __JSON_SEP) {                                   \
            ++str;                                                    \
            const char *begin = str, *end;                            \
            for (;;) {                                                \
                end = strchr(begin, __JSON_SEP);                      \
                if (*(end - 1) == '\\') {                             \
                    begin = end + 1;                                  \
                    continue;                                         \
                }                                                     \
                break;                                                \
            }                                                         \
            len      = (size_t)(end - str);                           \
            is_paren = 1;                                             \
        } else                                                        \
            len = strcspn(str, __JSON_SPACE __JSON_BEGIN __JSON_END); \
        str += len;                                                   \
        if (is_paren) {                                               \
            ++str;                                                    \
        }                                                             \
    }

static inline int
__hex_digit(const char c)
{
    if ('0' <= c && c <= '9')
        return c - '0';
    if ('a' <= c && c <= 'f')
        return c - 'a' + 10;
    if ('A' <= c && c <= 'F')
        return c - 'A' + 10;
    return -1;
}

static inline int
__hex_val(const char *c, uint32_t *result)
{
    const char *p;
    const size_t size = 4;
    int digit;
    *result = 0;

    for (p = c; (size_t)(p - c) < size; ++p) {
        *result <<= 4;
        digit = __hex_digit(*p);
        if (digit < 0 || digit > 15)
            return 1;
        *result |= (uint8_t)digit;
    }

    return 0;
}

/* sizeof(utf8) == 4 */
static struct {
    int32_t lo;
    int32_t hi;
} utf8[] = { { 0xd800, 0xdfff }, /* forbidden range: surrogates */
             { 0, 0x7f },
             { 0x80, 0x7ff },
             { 0x800, 0xffff },
             { 0x10000, 0x10ffff } };

static inline size_t
__encode_codepoint(char p[4], int32_t codepoint)
{
    size_t i, n = 0; /* number of bytes to be written */

    while (n < sizeof(utf8)) {
        if (utf8[n].lo <= codepoint && codepoint <= utf8[n].hi)
            break;
        n++;
    }

    if (n < 1 || n >= sizeof(utf8)) {
        /* Invalid! */
        return 0;
    }

    if (n == 1)
        p[0] = (char)codepoint;

    else {
        for (i = n - 1; i > 0; i--) {
            p[i] = (char)(0x80 | (0x3f & codepoint));
            codepoint >>= 6;
        }
        p[0] = (char)((0xf00 >> n) | codepoint);
    }

    return n;
}

static inline void
__decode_json(char *string, size_t len)
{
    char *offset       = string;
    size_t size        = 0;
    size_t written     = 0;
    uint32_t codepoint = 0;
    char *string_end   = string + len;

    while ((offset = strchr(offset, '\\'))) {
        switch (offset[1]) {
#define __case_single(c, sub) \
    case c:                   \
        size      = 2;        \
        written   = 1;        \
        offset[0] = sub;      \
        break;
            __case_single('"', '\"');
            __case_single('\\', '\\');
            __case_single('/', '/');
            __case_single('f', '\f');
            __case_single('b', '\b');
            __case_single('n', '\n');
            __case_single('r', '\r');
            __case_single('t', '\t');
#undef __case_single

        /* UTF-8 decode, 4 hex digits */
        case 'u':
            if (len < 4) {
                LOG_ERROR("JSON", "Invalid unicode string, abort decoding. String was: %s", string);
                return;
            }
            size = 6; /* \u + 4 hex digits */
            __hex_val(&offset[2], &codepoint);
            written = __encode_codepoint(offset, (int32_t)codepoint);
            break;

        /* Oups */
        default:
            LOG_ERROR("JSON", "Invalid json string, abort the decoding. String was: %s", string);
            return;
        }

        /* Now move the rest of the string of (size - written) characters*/
        if (written != size) {
            memmove(&offset[written], &offset[size],
                    (size_t)(string_end - &offset[size]) * sizeof(char));
            string_end -= (size - written);
            string_end[0] = '\0';
        }
    }
    string_end[0] = '\0';
}

LKT_PUSH
LKT_IGNORE_WARN_FMT_SECURITY
int
json_parse(const char *str, int asked_level, json_parse_callback call, void *user)
{
    int64_t level = 0;
    char key[LKT_LINE_MAX];
    char val[LKT_LINE_MAX];
    char tmp[LKT_LINE_MAX];

    for (;;) {
        size_t len = 0;
        __SKIP_JSON(str);

        /* Begin of a block */
        if ((len = strspn(str, __JSON_BEGIN))) {
            if (len >= INT64_MAX)
                return 1; /* Json error */
            level += (int64_t)len;
            str += len;
        }

        /* End of a block */
        else if ((len = strspn(str, __JSON_END))) {
            if ((level == asked_level) && call(NULL, NULL, 1, user) != 0)
                return 2;
            if (len >= INT64_MAX)
                return 1; /* Json error */
            str += len;
            level -= (int64_t)len;
        }

        /* Key Value */
        else {
            memset(key, 0, sizeof(key));
            memset(val, 0, sizeof(val));
            memset(tmp, 0, sizeof(tmp));

            __NEXT_JSON(str, len, key, tmp, LKT_LINE_MAX);
            __SKIP_JSON(str);
            __NEXT_JSON(str, len, val, tmp, LKT_LINE_MAX);

            if ((asked_level == level) && call(key, val, 0, user) != 0)
                return 2;
        }

        if (level <= 0)
            return 0;
    }
    LOG_UNREACHABLE;
}
LKT_POP

int
json_parse_get_count(const char *str, int asked_level)
{
    int64_t level = 0;
    int ret_len   = 0;

    for (;;) {
        size_t len = 0;
        __SKIP_JSON(str);

        /* Begin of a block */
        if ((len = strspn(str, __JSON_BEGIN))) {
            if (len >= INT64_MAX)
                return 1; /* Json error */
            level += (int64_t)len;
            str += len;
        }

        /* End of a block */
        else if ((len = strspn(str, __JSON_END))) {
            if (len >= INT64_MAX)
                return 1; /* Json error */
            ret_len += (level == asked_level);
            str += len;
            level -= (int64_t)len;
        }

        /* Key Value */
        else {
            __SKIP_NEXT_JSON(str, len);
            __SKIP_JSON(str);
            __SKIP_NEXT_JSON(str, len);
        }

        if (level <= 0)
            return ret_len;
    }
    LOG_UNREACHABLE;
}
