#[derive(Clone)]
pub enum Action {
    /// Resume playback from that kara.
    PlayFromKara,

    /// Pop the kara out of the queue.
    DeleteKaraFromQueue,

    /// Insert the kara in the queue. The kara will be inserted at the top of
    /// the queue, but after all the other inserted karas.
    InsertKaraInQueue,

    /// Add the kara at the end of the playlist.
    AddKaraToQueue,

    /// Add the kara to a playlist. The playlist will be selected by the user in
    /// a pop-up or something like that.
    AddKaraToPlaylist,

    /// Action to delete a kara from a playlist. The passed `u64` is the unique
    /// id of the playlist, the internal one.
    DeleteKaraFromPlaylist(u64),

    /// Open the playlist in a window for the user to browse.
    OpenPlaylist,

    /// Add all the content of a playlist to the queue.
    AddPlaylistToQueue,

    /// Insert all the content of a playlist in the queue.
    InsertPlaylistToQueue,

    /// Clear the content of the playlist, delete all the contained karas.
    ClearPlaylistContent,

    /// Connect to lektord.
    ConnectToLektord,

    /// Disconnect from lektord.
    DisconnectFromLektord,

    /// Refresh the list of playlists
    RefreshPlaylists,

    PlaybackPrevious,
    PlaybackPlay,
    PlaybackPause,
    PlaybackNext,
}

impl std::fmt::Debug for Action {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        use Action::*;
        match self {
            PlayFromKara => write!(f, "PlayFromKara"),
            DeleteKaraFromQueue => write!(f, "DeleteKaraFromQueue"),
            InsertKaraInQueue => write!(f, "InsertKaraInQueue"),
            AddKaraToQueue => write!(f, "AddKaraToQueue"),
            AddKaraToPlaylist => write!(f, "AddKaraToPlaylist"),
            DeleteKaraFromPlaylist(arg0) => {
                f.debug_tuple("DeleteKaraFromPlaylist").field(arg0).finish()
            }
            OpenPlaylist => write!(f, "OpenPlaylist"),
            AddPlaylistToQueue => write!(f, "AddPlaylistToQueue"),
            InsertPlaylistToQueue => write!(f, "InsertPlaylistToQueue"),
            ClearPlaylistContent => write!(f, "ClearPlaylistContent"),
            ConnectToLektord => write!(f, "ConnectToLektord"),
            DisconnectFromLektord => write!(f, "DisconnectFromLektord"),
            PlaybackPrevious => write!(f, "PlaybackPrevious"),
            PlaybackPlay => write!(f, "PlaybackPlay"),
            PlaybackPause => write!(f, "PlaybackPause"),
            PlaybackNext => write!(f, "PlaybackNext"),
            RefreshPlaylists => write!(f, "RefreshPlaylists"),
        }
    }
}

pub fn get_card_action_name(act: &Action) -> &'static str {
    use Action::*;
    match act {
        PlayFromKara => "Play from that kara",
        DeleteKaraFromQueue => "Delete from the queue",

        InsertKaraInQueue => "Insert in the queue",
        AddKaraToQueue => "Add to the queue",
        AddKaraToPlaylist => "Add to playlist",

        DeleteKaraFromPlaylist(_) => "Delete from playlist",
        OpenPlaylist => "Open playlist",
        AddPlaylistToQueue => "Add playlist to the queue",
        InsertPlaylistToQueue => "Insert the playlist in the queue",
        ClearPlaylistContent => "Clear the playlist content",

        ConnectToLektord => "Connect to lektord",
        DisconnectFromLektord => "Disconnect from lektord",

        PlaybackPrevious => "Previous",
        PlaybackPlay => "Play",
        PlaybackPause => "Pause",
        PlaybackNext => "Next",

        RefreshPlaylists => "Refresh playlists",
    }
}
